syntax = "proto3";

package soma.rpc;

import "soma/rpc/object_reference.proto";

// Summary of the fee for this transaction.
message TransactionFee {
  // base transaction fee.
  optional uint64 base_fee = 1;
  // Fee for each object operation (reads and writes).
  optional uint64 operation_fee = 2;
  // Fee based on transaction value.
  optional uint64 value_fee = 3;
  // Total fee deducted.
  optional uint64 total_fee = 4;
}