syntax = "proto3";

package soma.rpc;

// Reference to an object.
message ObjectReference {
  // The object id of this object.
  optional string object_id = 1;
  // The version of this object.
  optional uint64 version = 2;
  // The digest of this object.
  optional string digest = 3;
}